import ServiceMessage, { deserialize_ServiceMessage, serialize_ServiceMessage} from "./servicemessage";
import { IStream, IStreamCallback, Stream } from "./service_base";




export type { IStream, IStreamCallback };
export { Stream };

interface IServiceFunction {
   istream: boolean;
   input: (data: Uint8Array) => any;

   rstream: boolean;
   ret: (data: any) => Uint8Array;
}

export class Service<T> {
   public name: string = null as any;
   protected functions = new Map<string, IServiceFunction>();

   constructor() {}

   handleRequest(msg: ServiceMessage): IServiceFunction {
      const fnc = this.functions.get(msg.function);
      if (!fnc) throw new Error("Invalid function!");
      return fnc;
   }
}

type ISendMessageCB = (data: Uint8Array) => void;

export class ServiceProvider<T = any> {
   services = new Map<string, Service<T>>();
   addService(service: Service<T>) {
      this.services.set(service.name, service);
   }

   getSession(send: ISendMessageCB): Session<T> {
      return new Session(this, send);
   }
}

class Session<T> {
   ctx: Partial<T> = {};

   calls = new Map<
      string,
      {
         def: IServiceFunction;
         istream: Stream<any>;
      }
   >();
   constructor(
      private provider: ServiceProvider,
      private send: ISendMessageCB
   ) {}

   async close() {
      this.calls.forEach((call) => {
         call.istream.close();
         //TODO: Close rstreams or so...
      });
   }

   async onMessage(data: Uint8Array) {
      const msg = deserialize_ServiceMessage(data);
      try {
         if (this.calls.has(msg.id)) {
            const call = this.calls.get(msg.id);
            if (!call) throw new Error("This call does not exist anymore!");

            if (!msg.payload) {
               call.istream.close();
            } else {
               call.istream.send(call.def.input(msg.payload));
            }
         } else {
            const service = this.provider.services.get(msg.service);
            if (!service) throw new Error("Invalid Service!");

            const functionDefinition = service.handleRequest(msg);
            let input: any;
            if (functionDefinition.istream) {
               // Input is a stream
               const istream = new Stream<any>();
               input = istream;
               this.calls.set(msg.id, {
                  def: functionDefinition,
                  istream,
               });
               if (msg.payload) {
                  istream.send(functionDefinition.input(msg.payload));
               } else {
                  istream.close();
               }
            } else {
               input = functionDefinition.input(msg.payload);
            }

            const result: any = await (service as any)[msg.function](
               input,
               this.ctx
            );
            if (functionDefinition.rstream) {
               // Result is a stream
               (result as IStream<any>).subscribe((value, error, end) => {
                  if (!end) {
                     this.send(
                        serialize_ServiceMessage({
                           id: msg.id,
                           function: msg.function,
                           service: msg.service,
                           payload: error
                              ? undefined
                              : (functionDefinition.ret(value) as any),
                           error: error?.message as any,
                        })
                     );
                  } else {
                     this.send(
                        serialize_ServiceMessage({
                           id: msg.id,
                           function: msg.function,
                           service: msg.service,
                           payload: undefined as any,
                           error: undefined as any,
                        })
                     );
                  }
               });
            } else {
               this.send(
                  serialize_ServiceMessage({
                     id: msg.id,
                     function: msg.function,
                     service: msg.service,
                     payload: functionDefinition.ret(result),
                     error: undefined as any,
                  })
               );
            }
         }
      } catch (err) {
         if (this.calls.has(msg.id)) {
            this.calls.get(msg.id)?.istream.close();
         }
         this.send(
            serialize_ServiceMessage({
               id: msg.id,
               function: msg.function,
               service: msg.service,
               payload: undefined as any,
               error: err.message,
            })
         );
      }
   }
}
